
/* Chrysalide - Outil d'analyse de fichiers binaires
 * dex.c - équivalent Python du fichier "format/dex/dex.c"
 *
 * Copyright (C) 2012-2017 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */


#include "dex.h"


#include <pygobject.h>


#include <i18n.h>


#include <format/dex/class.h>
#include <format/dex/dex.h>


#include "../executable.h"
#include "../../analysis/content.h"
#include "../../helpers.h"



/* Crée un nouvel objet Python de type 'DexFormat'. */
static PyObject *py_dex_format_new(PyTypeObject *, PyObject *, PyObject *);

/* Dénombre le nombre de classes trouvées. */
static PyObject *py_dex_format_count_classes(PyObject *, PyObject *);

/* Fournit une classe du format chargée en mémoire. */
static PyObject *py_dex_format_get_class(PyObject *, PyObject *);



/******************************************************************************
*                                                                             *
*  Paramètres  : type = type de l'objet à instancier.                         *
*                args = arguments fournis à l'appel.                          *
*                kwds = arguments de type key=val fournis.                    *
*                                                                             *
*  Description : Crée un nouvel objet Python de type 'DexFormat'.             *
*                                                                             *
*  Retour      : Instance Python mise en place.                               *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_dex_format_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
{
    PyObject *result;                       /* Instance à retourner        */
    PyObject *content_obj;                  /* Objet pour le contenu       */
    PyObject *parent_obj;                   /* Objet pour le parent        */
    PyObject *status_obj;                   /* Objet pour la progression   */
    int ret;                                /* Bilan de lecture des args.  */
    GBinContent *content;                   /* Instance GLib du contenu    */
    GExeFormat *parent;                     /* Instance GLib du parent     */
    GtkStatusStack *status;                 /* Instance GTK de suivi       */
    GBinFormat *format;                     /* Création GLib à transmettre */

    ret = PyArg_ParseTuple(args, "OOO", &content_obj, &parent_obj, &status_obj);
    if (!ret) return NULL;

    ret = PyObject_IsInstance(content_obj, (PyObject *)get_python_binary_content_type());
    if (!ret)
    {
        PyErr_SetString(PyExc_TypeError, _("The first argument must be an instance of BinContent."));
        return NULL;
    }

    content = G_BIN_CONTENT(pygobject_get(content_obj));

    if (parent_obj == Py_None)
        parent = NULL;

    else
    {
        ret = PyObject_IsInstance(parent_obj, (PyObject *)get_python_executable_format_type());
        if (!ret)
        {
            PyErr_SetString(PyExc_TypeError, _("The second argument must be a container format or None."));
            return NULL;
        }

        parent = G_EXE_FORMAT(pygobject_get(parent_obj));

    }

    if (status_obj == Py_None)
        status = NULL;

    else
    {
        ret = PyObject_IsInstance(status_obj, (PyObject *)get_python_binary_content_type());
        if (!ret)
        {
            PyErr_SetString(PyExc_TypeError, _("The third argument must be a status bar object or None."));
            return NULL;
        }

        status = GTK_STATUS_STACK(pygobject_get(status_obj));

    }

    format = g_dex_format_new(content, parent, status);
    if (format == NULL)
    {
        PyErr_SetString(PyExc_RuntimeError, _("Unable to load the DEX format."));
        return NULL;
    }

    result = pygobject_new(G_OBJECT(format));

    g_object_unref(format);

    return (PyObject *)result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Dénombre le nombre de classes trouvées.                      *
*                                                                             *
*  Retour      : Quantité de classes présentes.                               *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_dex_format_count_classes(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    GDexFormat *format;                     /* Version native              */
    size_t count;                           /* Quantité à retourner        */

    format = G_DEX_FORMAT(pygobject_get(self));

    count = g_dex_format_count_classes(format);

    result = PyLong_FromLong(count);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : self = classe représentant un binaire.                       *
*                args = arguments fournis à l'appel.                          *
*                                                                             *
*  Description : Fournit une classe du format chargée en mémoire.             *
*                                                                             *
*  Retour      : Instance représentant une classe chargée.                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static PyObject *py_dex_format_get_class(PyObject *self, PyObject *args)
{
    PyObject *result;                       /* Trouvailles à retourner     */
    int ret;                                /* Bilan de lecture des args.  */
    int index;                              /* Indice de la classe         */
    GDexFormat *format;                     /* Version native              */
    GDexClass *class;                       /* Classe à communiquer        */

    ret = PyArg_ParseTuple(args, "i", &index);
    if (!ret) Py_RETURN_NONE;

    format = G_DEX_FORMAT(pygobject_get(self));

    class = g_dex_format_get_class(format, index);
    if (class == NULL) Py_RETURN_NONE;

    result = pygobject_new(G_OBJECT(class));

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Fournit un accès à une définition de type à diffuser.        *
*                                                                             *
*  Retour      : Définition d'objet pour Python.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

PyTypeObject *get_python_dex_format_type(void)
{
    static PyMethodDef py_dex_format_methods[] = {
        {
            "count_classes", (PyCFunction)py_dex_format_count_classes,
            METH_NOARGS,
            "count_classes($self, /)\n--\n\nCount the quantity of loaded Dex classes."
        },
        {
            "get_class", (PyCFunction)py_dex_format_get_class,
            METH_VARARGS,
            "get_class($self, index, /)\n--\n\nProvide a given loaded Dex class."
        },
        { NULL }
    };

    static PyGetSetDef py_dex_format_getseters[] = {
        { NULL }
    };

    static PyTypeObject py_dex_format_type = {

        PyVarObject_HEAD_INIT(NULL, 0)

        .tp_name        = "pychrysalide.format.dex.DexFormat",
        .tp_basicsize   = sizeof(PyGObject),

        .tp_flags       = Py_TPFLAGS_DEFAULT,

        .tp_doc         = "PyChrysalide Dex format",

        .tp_methods     = py_dex_format_methods,
        .tp_getset      = py_dex_format_getseters,
        .tp_new         = (newfunc)py_dex_format_new

    };

    return &py_dex_format_type;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : module = module dont la définition est à compléter.          *
*                                                                             *
*  Description : Prend en charge l'objet 'pychrysalide.format.dex.DexFormat'. *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool register_python_dex_format(PyObject *module)
{
    PyTypeObject *py_dex_format_type;       /* Type Python 'DexFormat'     */
    PyObject *dict;                         /* Dictionnaire du module      */

    py_dex_format_type = get_python_dex_format_type();

    dict = PyModule_GetDict(module);

    if (!register_class_for_pygobject(dict, G_TYPE_DEX_FORMAT,
                                  py_dex_format_type, get_python_executable_format_type()))
        return false;

    return true;

}
